#include <stdlib.h>
#include "terms.h"



union lpoint 
	{
	struct {
		Term value;
		union lpoint *nextp;
		List  nextl;
			} d;
	struct 	{
		List  l;
		union lpoint *next;
		} p;
	};
	

	
static union lpoint *l_next_free=NULL;	
static union lpoint *lbuffers[256];
static int l_blocks=0, l_alloc=0, l_free=0;

	
	
static union lpoint *list_pointer(List l)
	{
	register int bno,bpos;
	if(is_empty_list(l))
		return NULL;
	bno=l&0xff0000;
	bno/=0x10000;
	bpos=l&0xffff;
	return lbuffers[bno]+bpos;
	}	

static void lfree(List l, union lpoint *p)
	{
	p->p.next=l_next_free;
	p->p.l=l;
	l_next_free=p;
	l_free++;
	}
	
static union lpoint *lalloc(List *li)
	{
	union lpoint *p;
	if(l_next_free==NULL)
		{
		int i;
		if(l_blocks==256)
			{  puts("Internal error (too much lists)."); exit(0); }
		i=sizeof(union lpoint)*65500;
		p=lbuffers[l_blocks]=(union lpoint*)malloc(i);
		if(p==NULL)
			{  printf("Internal error (lack space for lists, %d blocks of %d bytes).\n",
					l_blocks, i); exit(0); }
		for(i=0;i<65500;i++)
			{
			p[i].p.next=l_next_free;
			l_next_free=p+i;
			p[i].p.l=0x7d000000+l_blocks*0x10000+i;
			}
		l_blocks++;
		}
	p=l_next_free;
	l_next_free=p->p.next;
	*li=p->p.l;
	p->d.value=0;
	p->d.nextp=NULL;
	p->d.nextl=0;
	l_alloc++;
	return p;
	}
	
	
List NewList(void)
	{
	return 0;
	}
	
int is_empty_list(List l)
	{
	return l==0;
	}
	
Term ListFirst(List l)
	{
	union lpoint *p;
	p=list_pointer(l);
	return p->d.value;
	}
	
List ListTail(List l)
	{
	union lpoint *p;
	p=list_pointer(l);
	return p->d.nextl;
	}	

List AppendFirst(List li,Term t)
	{
	union lpoint *p1, *p2;
	List l1;
	p1=lalloc(&l1);
	p1->d.value=t;
	if(li==0)
		return l1;
	p2=list_pointer(li);
	p1->d.nextl=li;
	p1->d.nextp=p2;
	return l1;
	}
	
List AppendLast(List li,Term t)
	{
	union lpoint *p1, *p2;
	List l1;
	p1=lalloc(&l1);
	p1->d.value=t;
	if(li==0)
		return l1;
	p2=list_pointer(li);
	while(p2->d.nextp!=NULL)
		p2=p2->d.nextp;
	p2->d.nextl=l1;
	p2->d.nextp=p1;
	return li;
	}
	
List ConcatList(List li, List l1)
	{
	union lpoint *p1, *p2;
	if(li==0)
		return l1;
	if(l1==0)
		return li;
	p1=list_pointer(l1);
	p2=list_pointer(li);
	while(p2->d.nextp!=NULL)
		p2=p2->d.nextp;
	p2->d.nextl=l1;
	p2->d.nextp=p1;
	return li;
	}

List ListSplit(List list, List cut, List *rest) 	
	{
	union lpoint *pi,*pcut, *pl;
	
	pi=list_pointer(list);
	pcut=list_pointer(cut);
	
	if(list==cut)
		{
		*rest=pi->d.nextl;
		pi->d.nextp=NULL;
		pi->d.nextl=0;
		return 0;
		}
		
	pl=pi;
	while(pl->d.nextl!=cut)
		pl=pl->d.nextp;
	pcut=pl->d.nextp;
	pl->d.nextp=NULL;
	pl->d.nextl=0;
	*rest=pcut->d.nextl;
	pcut->d.nextp=NULL;
	pcut->d.nextl=0;
	return list;
	}
	
void FreeList(List li)
	{
	union lpoint *l,*l1;
	List li1;
	l=list_pointer(li);
	while(l!=NULL)
		{
		l1=l->d.nextp;
		li1=l->d.nextl;
		FreeAtomic(l->d.value);
		lfree(li,l);
		l=l1;
		li=li1;
		}
	}
	
void RemoveList(List li)
	{
	union lpoint *l,*l1;
	List li1;
	l=list_pointer(li);
	while(l!=NULL)
		{
		l1=l->d.nextp;
		li1=l->d.nextl;
		lfree(li,l);
		l=l1;
		li=li1;
		}
	}	
	

int ListLength(List l)
	{
	union lpoint *p;
	int len;
	len=0;
	if(l==0) return 0;
	p=list_pointer(l);
	while(p!=NULL)
		{
		p=p->d.nextp;
		len++;
		}
	return len;
	}
	
Term ListNth(List l, int arg)
	{
	union lpoint *p;
	p=list_pointer(l);
	while(arg>1)
		{
		p=p->d.nextp;
		arg--;
		}
	return p->d.value;
	}
	
List ListNthList(List l, int arg)
	{
	union lpoint *p;
	if(arg<2)
		return l;
	p=list_pointer(l);
	while(arg>2)
		{
		p=p->d.nextp;
		arg--;
		}
	return p->d.nextl;
	}


List CutFromList(List list, List cut)
     {
     union lpoint *pi,*pcut, *pl;
     List rest;
	pi=list_pointer(list);
	pcut=list_pointer(cut);

	if(list==cut)
		{
		rest=pi->d.nextl;
		pi->d.nextp=NULL;
		pi->d.nextl=0;
        FreeList(list);
		return rest;
		}
		
	pl=pi;
	while(pl->d.nextl!=cut)
		pl=pl->d.nextp;
	pcut=pl->d.nextp;
	pl->d.nextp=pcut->d.nextp;
	pl->d.nextl=pcut->d.nextl;

	pcut->d.nextp=NULL;
	pcut->d.nextl=0;
    FreeList(cut);
	return list;
	}


void InsertList(List l,Term ins)
	{
	union lpoint *p,*p1;
	List l1;
	p=list_pointer(l);
	p1=lalloc(&l1);
	p1->d.value=ins;
	p1->d.nextl=p->d.nextl;
	p1->d.nextp=p->d.nextp;
	p->d.nextl=l1;
	p->d.nextp=p1;
	}
	

void ChangeList(List li, Term nv)
     {
     union lpoint *p;
     p=list_pointer(li);
     /*FreeAtomic(p->d.value);*/
     p->d.value=nv;
     }




void ListStatistics(void)
	{
	printf("Lists: %d blocks, %d allocs, %d frees ( %d remain)\n",
		l_blocks,l_alloc,l_free,l_alloc-l_free);
	}

void ListStat1(void)
	{
	printf("%d lists (%d blocks)\n",l_alloc-l_free,l_blocks);
	}





void DumpList(List l)
	{
	printf("\t[\n\t");
	while(!is_empty_list(l))
		{
		WriteTerm(ListFirst(l));
		printf("\n\t");
		l=ListTail(l);
		}
	printf("]\n");
	}


int ListMember(List l, Term t)
	{
	int i=1;
	while(!is_empty_list(l))
		{
		if(ListFirst(l)==t)
			return i;
		i++;
		l=ListTail(l);
		}
	return 0;
	}

List IntersectList(List l1, List l2)
	{
	List li;
	li=NewList();
	while(!is_empty_list(l1))
		{
		if(ListMember(l2,ListFirst(l1)))
			li=AppendLast(li,CopyTerm(ListFirst(l1)));
		l1=ListTail(l1);
		}
	return li;
	}


List SortedList(List li, int (*cmp)(Term,Term))
	{
	List l,l1,l2,li1;
	li1=li;
	if(is_empty_list(li))
		return NewList();
	l=AppendFirst(NewList(),ListFirst(li));
	li=ListTail(li);
	while(!is_empty_list(li))
		{
		Term t;
		t=ListFirst(li);
		if(cmp(t,ListFirst(l))<=0)
			{
			l=AppendFirst(l,t);
			goto cnt;
			}
		l1=l;
		l2=ListTail(l);
		while(!is_empty_list(l2))
			{
			if(cmp(t,ListFirst(l2))<=0)
				{
				InsertList(l1,t);
				goto cnt;
				}
			l1=l2;
			l2=ListTail(l2);
			}
		InsertList(l1,t);
	cnt:
		li=ListTail(li);
		}
	RemoveList(li1);
	return l;
	}
	
List MakeList2(Term t1, Term t2)
	{
	register List l;
	l=AppendFirst(NewList(),t2);
	l=AppendFirst(l,t1);
	return l;
	}
	
List MakeList3(Term t1, Term t2, Term t3)
	{
	register List l;
	l=AppendFirst(NewList(),t3);
	l=AppendFirst(l,t2);
	l=AppendFirst(l,t1);
	return l;
	}
	
List MakeList4(Term t1, Term t2, Term t3, Term t4)
	{
	register List l;
	l=AppendFirst(NewList(),t4);
	l=AppendFirst(l,t3);
	l=AppendFirst(l,t2);
	l=AppendFirst(l,t1);
	return l;
	}
		

	
/*
static char cbuf[100];


static int t_cmp(Term t1, Term t2)
	{
	char c1,c2;
	c1=AtomicType(t1);
	c2=AtomicType(t2);
	if(c1<c2) return -1;
	if(c1>c2) return 1;
	switch(c1)
		{
		case 'i':
			if(t1==t2) return 0;
			if(IntegerValue(t1)<IntegerValue(t2))
				return -1;
			return 1;
		case 'a':
			return strcmp(AtomValue(t1),AtomValue(t2));
		case 'f':
			if(FloatValue(t1)==FloatValue(t2)) return 0;
			if(FloatValue(t1)<FloatValue(t2))
				return -1;
			return 1;
		default: return 0;
		}
	return 0;
	}	
	
static void displlist(List l)
	{
	Atomic last_a=0;
	int reps=0;
	while(!is_empty_list(l))
		{
		Atomic a;
		a=ListFirst(l);
		if(last_a==0)
			{
			last_a=a;
			reps=1;
			l=ListTail(l);
			continue;
			}
		if(a==last_a)
			{
			reps++;
			l=ListTail(l);
			continue;
			}
		DisplayTerm(last_a);
		printf("\t\t:%d\n",reps);
		reps=1;
		last_a=a;
		l=ListTail(l);
		}
	
	DisplayTerm(last_a);
	printf("\t\t:%d\n",reps);
	}

*/
